/* $Id$ 
 *
 * stddef.h
 */

#ifndef _STDDEF_H
#define _STDDEF_H

typedef signed int   ptrdiff_t;
typedef unsigned int size_t;
typedef signed int   ssize_t;

#ifdef __cplusplus
#define NULL 0
#else
#define NULL ((void *)0)
#endif

#define offsetof(__t,__m) ((size_t)&(((__t*)0)->__m))

#endif /* _STDDEF_H */

